//////////////////////////////////////////////////////////////////////////////
//   ///////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include "_DRAW_3D.h"
#include "_3D.h"

//////////////////////////////////////////////////////////////////////////////
//   /////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void Falloff(int f)
{ 
  DrawRect(RegionFull(),RGBis(f,f,f),modeMUL);
};

//////////////////////////////////////////////////////////////////////////////
inline tfloat dist(tfloat x1, tfloat y1, tfloat x2, tfloat y2)
{
  x2 = x2 - x1; y2 = y2 - y1;
  return sqrt(x2 * x2 + y2 * y2);
};

//////////////////////////////////////////////////////////////////////////////
inline dwRGB DIColor(dwRGB c, int m = 255)
{ 
  return RGBmultV(c, DrawIntensityI * m / 255);
};

//////////////////////////////////////////////////////////////////////////////
// () GArray //////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
class GArray
{ public:
  int xD, yD, vFilter, Mode; 
  tfloat vLevel, vMin, vMax, vOver, vLimit, vRandom, iK1, iK2, iDeep, sPaint, sFalloff;
  tfloat_arr *Value, *Speed;
  dwRGB ColorD, ColorL;

  tfloat (*EPaintFunc)(tfloat);

  int gx(int x) { return (xD + x) % xD; };
  int gy(int y) { return (yD + y) % yD; };

  void filterUp()                             //  : 
  {
    for (int y = 0; y < yD; y++)          //      
    for (int x = 0; x < xD; x++)
    { (*Speed)[y * xD + x] = (*Speed)[gy(y + 1) * xD + x] + fRandom(-vRandom, vRandom);
     (*Value)[y * xD + x] = (*Value)[gy(y + 1) * xD + x];
    };
  };

  void filterDown()                           //  : 
  {
    for (int y = yD - 1; y >= 0; y--)
    for (int x = 0; x < xD; x++) 
    { (*Speed)[y * xD + x] = (*Speed)[gy(y - 1) * xD + x] + fRandom(-vRandom, vRandom); 
     (*Value)[y * xD + x] = (*Value)[gy(y - 1) * xD + x];
    };
  };

  void filterRight()                           //  : 
  {
    for (int y = 0; y < yD; y++) 
    for (int x = xD - 1; x >= 0; x--)
    { (*Speed)[y * xD + x] = (*Speed)[y * xD + gx(x - 1)] + fRandom(-vRandom, vRandom);
     (*Value)[y * xD + x] = (*Value)[y * xD + gx(x - 1)];
    };
  };

  void filterLeft()                           //  : 
  {
    for (int y = 0; y < yD; y++)
    for (int x = 0; x < xD; x++)
    { (*Speed)[y * xD + x] = (*Speed)[y * xD + gx(x + 1)] + fRandom(-vRandom, vRandom);
     (*Value)[y * xD + x] = (*Value)[y * xD + gx(x + 1)];
    };
  };

  void Filter()                              //   []
  {
    switch(vFilter)
    { default: break;
      case 1: filterUp(); break;
      case 2: filterDown(); break;
      case 3: filterRight(); break;
      case 4: filterLeft(); break;
    };
  };

  void SetColors(dwRGB cd, dwRGB cl) { ColorD = cd; ColorL = cl; };

  void SetFilter(int f)  { vFilter = f; };
 
  void SetMinMax(tfloat mn, tfloat mx, tfloat vl = 0.25, tfloat vo = -5) {vMin = mn; vMax = mx, vLimit = vl; vOver = vo; };

  void Create()                                             //  
  {
    Done();  xD = DrawXR / 8; yD = DrawYR  / 8;              //    -
    Value = (tfloat_arr *)malloc(xD * yD * sizeof(tfloat)); //  
    Speed = (tfloat_arr *)malloc(xD * yD * sizeof(tfloat));
    for (int y = 0; y < yD; y++)
    for (int x = 0; x < xD; x++)                            //     
    {(*Value)[y * xD + x] = vLevel;(*Speed)[y * xD + x] = 0;
    };
  };

  void Done() { if (Value) free(Value); if (Speed) free(Speed); Speed = Value = NULL; };

  GArray()                                   // 
  {
    Mode = modeADD; vFilter = 0; SetColors(RGBis(0,125,255), RGBis(255,145,0));
    vFilter = 0; vLevel = 0; SetMinMax(-3.75, 3.75, 0.15);  Value = Speed = NULL; vRandom = 0.075;
    iK1 = iK2 = 0.025; iDeep = 7; sPaint = 0.75; sFalloff = 0.97;
  };

  ~GArray() { Done(); };                     //  

  void Paint(int x, int y, int rad)          // 
  { 
    tfloat r;
    int xx = (x - 4) / 8,   yy = (y - 4) / 8;
    for (int y = yy - rad; y <= yy + rad; y++)
    for (int x = xx - rad; x <= xx + rad; x++)
    if ((r = dist(xx, yy, x, y)) <= rad) floatLinc((*Speed)[gy(y) * xD + gx(x)], vMax,(1 - Hermit(r / rad)) * sPaint);
  };

  void Draw()                                // 
  {
    tfloat v, r, vD = vMax - vMin;
    int d, x, y, ln, lnx;
    for (y = 0; y < yD; y++)                                            //        : 
    { ln = y * xD;                                                      //  x-
      for (x = 0; x < xD; x++)
      { lnx = ln + x;(*Speed)[lnx] -= iK1 * ((v = (*Value)[lnx]) - vLevel); //   
        for (d = 1; d <= iDeep; d++)                                    //      - iDeep
       (*Speed)[lnx] += iK2 * ((*Value)[ln + gx(x - d)] + (*Value)[ln + gx(x + d)] - 2 * v) / d;
      };
    };
    for (x = 0; x < xD; x++)                                            //        : 
    { for (y = 0; y < yD; y++)
      { lnx = y * xD + x; (*Speed)[lnx] -= iK1 * ((v = (*Value)[lnx]) - vLevel); //   
        for (d = 1; d <= iDeep; d++)                                    //      - iDeep
       (*Speed)[lnx] += iK2 * ((*Value)[xD * gy(y - d) + x] + (*Value)[xD * gy(y + d) + x] - 2 * v) / d;
      };
    };

    for (int y = 0; y < yD; y++)
    for (int x = 0; x < xD; x++)
    { lnx = y * xD + x;  v = (*Value)[lnx] += (*Speed)[lnx] *= sFalloff;
      if ((r = max(min(v, vMax), vMin)) >= vOver)
      { DrawCircle(4 + x * 8, 4 + y * 8, max(abs(r), vLimit), DIColor(RGBblend(ColorD, ColorL,(r - vMin) / vD * 255 + 0.5)), Mode); // :    
        if ((r = abs(r)) > 2.75) DrawCircle(4 + x * 8, 4 + y * 8, r - 2.75, DIColor(0xffffffff, 175), modeADD);
      };

    };  
  };
};

//////////////////////////////////////////////////////////////////////////////
//    /////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
GArray garray;

//////////////////////////////////////////////////////////////////////////////////////////////////
int MousePressNow = no;

void mouseLBDown()                           //  []  
{ 
  MousePressNow = Yes;
};

void mouseLBUp()                           //  []  
{ 
  MousePressNow = no;
};

//////////////////////////////////////////////////////////////////////////////////////////////////
int keyExit = 0, HelpTimer = 0, BrushSize = 17, ColorI = 0;

void KeysLook(int force = no)                   //   
{
  if (!force && !II.AnyKey) return;

  if (II.KeyUp[_Enter_]) ColorI = (ColorI + 1) % 3;  
  if (II.KeyUp[_Quote_]) ColorI = (3 + ColorI - 1) % 3;
  switch(ColorI)
  { case 0: garray.SetColors(RGBis(0,125,255), RGBis(255,145,0)); break;
    case 1: garray.SetColors(RGBis(0,200,255), RGBis(175,225,0)); break; 
    case 2: garray.SetColors(RGBis(185,0,70), RGBis(255,110,0)); break; 
  };


  if (II.KeyUp[_1_]) BrushSize = 7;
  if (II.KeyUp[_2_]) BrushSize = 11; 
  if (II.KeyUp[_3_]) BrushSize = 17; 
  if (II.KeyUp[_4_]) BrushSize = 25;

  if (II.KeyUp[_7_]) garray.sPaint = 0.25;
  if (II.KeyUp[_8_]) garray.sPaint = 0.75; 
  if (II.KeyUp[_9_]) garray.sPaint = 1.25; 
  if (II.KeyUp[_0_]) garray.sPaint = 2;

  if (II.KeyUp[_CU_]) garray.SetFilter(1);
  if (II.KeyUp[_CD_]) garray.SetFilter(2); 
  if (II.KeyUp[_CR_]) garray.SetFilter(3); 
  if (II.KeyUp[_CL_]) garray.SetFilter(4);
  if (II.KeyUp[_Space_]) garray.SetFilter(0);
  if (II.KeyUp[_Del_]) if (garray.vRandom > 0.001) garray.vRandom = 0; else garray.vRandom = 0.075; 
  if (II.KeyUp[_F1_]) if (HelpTimer) HelpTimer = 0; else HelpTimer = 25 * 10;
  keyExit = (II.KeyUp[_Esc_]);
  II.Reset();
};


//////////////////////////////////////////////////////////////////////////////
//     /////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
TextSPR *info;

void CommonsInit()
{ RandomSeed();

  FC_FpsSet(AUTO);            //    
  FC_DrawPaint = Yes;         //  fps   DrawPaint();

  Draw2D_Init();
  ForceActivate();
  ScreenLoadWait();

  info = new TextSPR(L"TYPE_EN_RU");

  garray.Create(); //garray.vRadDrawLim = 0.5;

  OnMouseLButtonDownDo(mouseLBDown);   //     
  OnMouseLButtonUpDo(mouseLBUp);   //     
//  OnKeyUpDo(keyPressed);       //     
  KeysLook(Yes);                //     ...
};


//////////////////////////////////////////////////////////////////////////////////////////////////
void CommonsDone()
{
  delete info;
  Draw2D_Done();
};

//////////////////////////////////////////////////////////////////////////////////////////////////
void DrawInfo()
{
  tRegion r = RegionCenterIs(208, 288);
//  tRegion r = RegionCenterIs(1000, 500);
  DrawRegionStore.Save(r);
  DrawRect(r, RGBis(50, 50, 50), modeMUL); 
  info->Place(0, 0, modeADD + fromCC, "Conrol Keys:\n\n\nQutes, Enter - change Color\n...............................................................................\n\n1, 2, 3, 4 - set Brush Size\n\n7, 8, 9, 0 - set Brush Force\n\nCursor - set Move Direction\n\nSpace - set No Move\n\nDel - set On/Off Randomize\n...............................................................................\n\nF1 - hide/show Help\n\nEsc - Exit\n...............................................................................\n\n\nAnd... - Mouse Click & Motion"); info->Draw();
};

//////////////////////////////////////////////////////////////////////////////////////////////////
void DrawAll()
{
  if (MousePressNow) garray.Paint(II.Mouse.X, II.Mouse.Y, BrushSize);
  garray.Filter();
  garray.Draw();               // 
  if (HelpTimer) DrawInfo();
  DrawPaint();
  if (HelpTimer) { DrawRegionStore.Restore();  HelpTimer--; };
};

//////////////////////////////////////////////////////////////////////////////
//   //////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
void Main()
{
  CommonsInit();                  //   ;  

  SetDrawIntensity(0.5);          //      
  //SetCursor(LoadCursor(NULL,IDC_CROSS));
  //SetCursor(NULL);

  while (!keyExit)
   { Falloff(120);
     KeysLook();
     DrawAll();
   };   

  CommonsDone();                  //   ;
};
//////////////////////////////////////////////////////////////////////////////
int LetProcess()
{
  WinName = L"CG : Array Wave"; //    
  VS_CColor = RGBis(0,0,0);    //   
  DrawSet(FullScreen);         //   ( )  , FullScreen   
  //SetDraw(640,480);
  ProcessFunc = Main;           //    -  
  return on;
};
int Init_LetProcess = LetProcess();

